#include <array>
#include <memory>

#include <Arduino.h>
#include <WiFi.h>
#include <WiFiMulti.h>
#include <SPIFFS.h>

#include "WebServer.h"
#include "Fader.h"
#include "UDPProto.h"
#include "UpdateServer.h"
#include "Config.h"

#include "Animation/AllAnimations.h"
#include "Animation/AnimationController.h"

#include <esp32_digital_led_lib.h>
#include <esp32_digital_led_funcs.h>

#include "coreids.h"

const uint32_t FRAME_INTERVAL_US = 16666;
const uint32_t NUM_STRIPS = 8;
const uint32_t NUM_LEDS = 16;
const uint32_t FLIP_STRIPS_MASK = 0x000000AA;

std::array<strand_t, 1> STRANDS { // Avoid using any of the strapping pins on the ESP32, anything >=32, 16, 17... not much left.
  strand_t {.rmtChannel = 0, .gpioNum = 4, .ledType = LED_SK6812W_V1, .brightLimit = 32, .numPixels = NUM_LEDS * NUM_STRIPS},
};

bool led_on = false;

size_t led_idx;
WiFiMulti wiFiMulti;

Fader ledFader(NUM_STRIPS, NUM_LEDS, 1, FLIP_STRIPS_MASK);
UDPProto udpProto(&ledFader);
UpdateServer *updateServer;

AnimationController animController(&ledFader);

bool initLEDs()
{
  /****************************************************************************
     If you have multiple strands connected, but not all are in use, the
     GPIO power-on defaults for the unused strand data lines will typically be
     high-impedance. Unless you are pulling the data lines high or low via a
     resistor, this will lead to noise on those unused but connected channels
     and unwanted driving of those unallocated strands.
     This optional gpioSetup() code helps avoid that problem programmatically.
  ****************************************************************************/

  digitalLeds_initDriver();

  for (int i = 0; i < STRANDS.size(); i++) {
    gpioSetup(STRANDS[i].gpioNum, OUTPUT, LOW);
  }

  strand_t *strands[STRANDS.size()];
  for (int i = 0; i < STRANDS.size(); i++)
  {
    strands[i] = &STRANDS[i];
  }
  int rc = digitalLeds_addStrands(strands, STRANDS.size());
  if (rc)
  {
    Serial.print("Init rc = ");
    Serial.println(rc);
    return false;
  }

  for (int i = 0; i < STRANDS.size(); i++)
  {
    strand_t *pStrand = strands[i];
    Serial.print("Strand ");
    Serial.print(i);
    Serial.print(" = ");
    Serial.print((uint32_t)(pStrand->pixels), HEX);
    Serial.println();
#if DEBUG_ESP32_DIGITAL_LED_LIB
    dumpDebugBuffer(-2, digitalLeds_debugBuffer);
#endif
  }

  led_idx = 0;
  return true;
}

enum LEDState {
  UDP,
  ANIMATION,
  TRANSITION,
  TRANSITION_FADE2BLACK
};

static void ledFSM(void)
{
  static bool stateEntered = true;
  static LEDState ledState = ANIMATION;
  static LEDState nextState;

  static unsigned long lastUDPUpdate = 0;

  LEDState lastState = ledState;

  switch(ledState) {
    case ANIMATION:
      Serial.print("A");

      if(stateEntered) {
        animController.restart();
      }

      animController.loop();

      if(animController.isIdle()) {
        animController.changeAnimation(std::unique_ptr<Animation>(new FireAnimation(&ledFader, false)));
      }

      if(((WiFi.status() == WL_CONNECTED) || (WiFi.getMode() == WIFI_MODE_AP)) &&
          udpProto.check()) {
        // UDP packet received -> transition to UDP state
        nextState = UDP;
        ledState = TRANSITION;
      }
      break;

    case UDP:
      Serial.print("U");
      if((WiFi.status() != WL_CONNECTED) && (WiFi.getMode() != WIFI_MODE_AP)) {
        // WiFi disconnected -> switch to animation immediately to show reconnect animation
        ledState = ANIMATION;
        break;
      }

      if(udpProto.loop()) {
        lastUDPUpdate = millis();
      } else if(millis() - lastUDPUpdate > 3000) {
        // seems like no more UDP data arrives -> transition to ANIMATION state
        nextState = ANIMATION;
        ledState = TRANSITION;
      }

      ledFader.update();
      break;

    case TRANSITION:
      Serial.print("S");
      if(stateEntered) {
        animController.stop();
      }

      if(animController.isIdle()) {
        ledState = TRANSITION_FADE2BLACK;
      } else {
        animController.loop();
      }
      break;

    case TRANSITION_FADE2BLACK:
      Serial.print("F");
      if(stateEntered) {
        ledFader.set_fadestep(2);
        ledFader.fade_color(Fader::Color{0, 0, 0, 0});
      }

      ledFader.update();
      if(!ledFader.something_changed()) {
        // transition finished
        ledState = nextState;
        lastUDPUpdate = millis();
      }

      break;
  }

  stateEntered = (lastState != ledState);
}

static void ledTask( void * parameter )
{
  uint64_t frame = 0;

  /* loop forever */
  for(;;){

    uint32_t start_time = micros();

    ledFSM();

    led_idx++;
    if(led_idx >= STRANDS[0].numPixels) {
      led_idx = 0;
    }

    strand_t *strands[STRANDS.size()];
    for (int i = 0; i < STRANDS.size(); i++)
    {
      strands[i] = &STRANDS[i];
    }

    if(ledFader.something_changed()) {
      const std::vector<Fader::Color> &colors = ledFader.get_color_values();

      for (size_t i = 0; i < STRANDS[0].numPixels; i++) {
        const Fader::Color &c = colors[i];
        STRANDS[0].pixels[i] = pixelFromRGBW(c.r, c.g, c.b, c.w);
      }
    }

    digitalLeds_drawPixels(strands, 1);

    frame++;

    uint32_t now = micros();
    uint32_t duration = now - start_time;

    if(frame % 50 == 0) {
      uint32_t load = 100 * duration / FRAME_INTERVAL_US;

      Serial.print("CPU Load: ");
      Serial.print(load);
      Serial.println(" %");
    }

    if(duration < FRAME_INTERVAL_US) {
      delayMicroseconds(FRAME_INTERVAL_US - duration);
    }
  }
  /* delete a task when finish, 
  this will never happen because this is infinity loop */
  vTaskDelete( NULL );
}

void wifi_setup(void)
{
  Serial.println("Trying to connect...");

  for(size_t tries = 0; tries < 10; tries++)
  {
    if(wiFiMulti.run() == WL_CONNECTED) {
      animController.changeAnimation(std::unique_ptr<Animation>(new ConnectionEstablishedAnimation(&ledFader, true)));

      Serial.println("");
      Serial.println("WiFi connected");
      Serial.println("IP address: ");
      Serial.println(WiFi.localIP());
      break;
    }

    led_on = !led_on;
    digitalWrite(2, led_on);
    delay(100);

    Serial.print(".");
  }

  if(WiFi.status() != WL_CONNECTED) {
    Serial.println("Connection failed, setting up access point...");

    IPAddress apIP(192, 168, 42, 1);

    WiFi.mode(WIFI_AP);
    WiFi.softAPConfig(apIP, apIP, IPAddress(255, 255, 255, 0));
    WiFi.softAP("🕯️💡☀️", "Licht234");
    WiFi.enableAP(true);

    animController.changeAnimation(std::unique_ptr<Animation>(new ConnectionEstablishedAnimation(&ledFader, false)));
  }

}

void setup()
{
  pinMode(2, OUTPUT); // On-Board LED

  Serial.begin(115200);
  delay(10);

  if(!initLEDs()) {
    Serial.println("LED setup failed!");
    while(true) {
      delay(100);
    }
  }

  if(!SPIFFS.begin()) {
    Serial.println("SPIFFS setup failed!");
    while(true) {
      delay(100);
    }
  }

  // load configuration (especially crypto data)
  Config::instance().load();

  digitalWrite(2, HIGH);

  Serial.println();
  Serial.println();

  animController.changeAnimation(std::unique_ptr<Animation>(new ConnectingAnimation(&ledFader)), false);

  xTaskCreatePinnedToCore(
      ledTask,           /* Task function. */
      "LED Task",        /* name of task. */
      10000,                    /* Stack size of task */
      NULL,                     /* parameter of the task */
      3,                        /* priority of the task */
      NULL,                    /* Task handle to keep track of created task */
      CORE_ID_LED);

  // Connect the WiFi network (or start an AP if that doesn't work)
  for (auto &net : Config::instance().getWLANList())
  {
    Serial.print("Adding network ");
    Serial.println(net.ssid.c_str());
    wiFiMulti.addAP(net.ssid.c_str(), net.password.c_str());
  }

  wifi_setup();

  // start the UDP server
  udpProto.start(2703);

  // start the web server
  WebServer::instance().setFader(&ledFader);
  WebServer::instance().setAnimationController(&animController);
  WebServer::instance().start();

  // start the update server
  updateServer = new UpdateServer(Config::instance().getCRPassword());
  updateServer->start();
}

void loop() {
  led_on = !led_on;
  digitalWrite(2, led_on);
  delay(500);

  // reconnect WiFi when connection is lost
  if(WiFi.status() == WL_CONNECTION_LOST) {
    Serial.println("WLAN disconnected. Restarting setup.");
    wifi_setup();
  }
}
